from scapy.fields import (
    BitEnumField,
    BitField,
    ByteField,
    ConditionalField,
    FlagsField,
    LongField,
    ShortField,
    SignedByteField,
    XBitField,
    XByteField,
    XIntField,
    XStrFixedLenField,
)
from scapy.layers.l2 import Ether
from scapy.packet import Packet, bind_layers

from .fields import PortIdentityField, TimestampField


class PTPv2(Packet):
    name = "PTPv2"

    MSG_TYPES = {
        0x0: "Sync",
        0x2: "PdelayReqest",
        0x3: "PdelayResponse",
        0x8: "FollowUp",
        0xA: "PdelayResponseFollowUp",
    }

    FLAGS = [
        "LI61",
        "LI59",
        "UTC_REASONABLE",
        "TIMESCALE",
        "TIME_TRACEABLE",
        "FREQUENCY_TRACEABLE",
        "?",
        "?",
        "ALTERNATE_MASTER",
        "TWO_STEP",
        "UNICAST",
        "?",
        "?",
        "profileSpecific1",
        "profileSpecific2",
        "SECURITY",
    ]

    fields_desc = [
        BitField("majorSdoId", 1, 4),
        BitEnumField("messageType", 0, 4, MSG_TYPES),
        XBitField("minorVersionPTP", 0, 4),
        BitField("versionPTP", 0x2, 4),
        ShortField("messageLength", 34),
        ByteField("domainNumber", 0),
        XByteField("minorSdoId", 0),
        FlagsField("flags", 0, 16, FLAGS),
        LongField("correctionField", 0),
        XIntField("messageTypeSpecific", 0),
        PortIdentityField("sourcePortIdentity", None),
        ShortField("sequenceId", 0),
        XByteField("controlField", 0),
        SignedByteField("logMessageInterval", -3),
        # Sync (twoStep flag set) or PdelayReq
        ConditionalField(
            BitField("reserved", 0, 80),
            lambda pkt: (pkt.is_sync and pkt.has_twostepflag_set) or pkt.is_pdelay_req,
        ),
        # Sync (twoStep flag not set)
        ConditionalField(
            TimestampField("originTimestamp", 0),
            lambda pkt: pkt.is_sync and not pkt.has_twostepflag_set,
        ),
        # FollowUp
        ConditionalField(
            TimestampField("preciseOriginTimestamp", 0), lambda pkt: pkt.is_followup
        ),
        # Sync (twoStep flag not set) or FollowUp
        ConditionalField(
            XStrFixedLenField("informationTlv", bytes(32), 32),
            lambda pkt: (pkt.is_sync and not pkt.has_twostepflag_set)
            or pkt.is_followup,
        ),
        # PdelayReq
        ConditionalField(BitField("reserved2", 0, 80), lambda pkt: pkt.is_pdelay_req),
        # PdelayResp
        ConditionalField(
            TimestampField("requestReceiptTimestamp", 0), lambda pkt: pkt.is_pdelay_resp
        ),
        # PdelayRespFollowUp
        ConditionalField(
            TimestampField("responseOriginTimestamp", 0),
            lambda pkt: pkt.is_pdelay_resp_followup,
        ),
        # PdelayResp or PdelayRespFollowUp
        ConditionalField(
            PortIdentityField("requestingPortIdentity", 0),
            lambda pkt: pkt.is_pdelay_resp or pkt.is_pdelay_resp_followup,
        ),
    ]

    @property
    def is_sync(self):
        return self.messageType == 0x0

    @property
    def is_followup(self):
        return self.messageType == 0x8

    @property
    def is_pdelay_req(self):
        return self.messageType == 0x2

    @property
    def is_pdelay_resp(self):
        return self.messageType == 0x3

    @property
    def is_pdelay_resp_followup(self):
        return self.messageType == 0xA

    @property
    def has_twostepflag_set(self):
        return "TWO_STEP" in self.flags

    def extract_padding(self, s):
        return "", s


bind_layers(Ether, PTPv2, type=0x88F7)



import struct
from scapy.fields import XStrFixedLenField, BitField


class TimestampField(BitField):
    def __init__(self, name, default):
        BitField.__init__(self, name, default, 80)

    def any2i(self, pkt, val):
        if val is None:
            return val
        ival = int(val)
        fract = int((val - ival) * 1e9)
        return (ival << 32) | fract

    def i2h(self, pkt, val):
        int_part = val >> 32
        frac_part = val & (1 << 32) - 1
        return int_part + frac_part * 1e-9


class PortIdentityField(XStrFixedLenField):
    encoding = "!BBBBBBBBh"

    @classmethod
    def from_mac(cls, mac, port):
        mac_bytes = mac.split(":")
        if len(mac_bytes) != 6:
            raise ValueError("Invalid MAC Address")

        return struct.pack(
            cls.encoding,
            int(mac_bytes[0], 16),
            int(mac_bytes[1], 16),
            int(mac_bytes[2], 16),
            0xFF,
            0xFE,
            int(mac_bytes[3], 16),
            int(mac_bytes[4], 16),
            int(mac_bytes[5], 16),
            port,
        )

    def __init__(self, name, default):
        XStrFixedLenField.__init__(self, name, default, length=10)

    def i2h(self, pkt, val):
        if val is None:
            return "None"
        p = struct.unpack(self.encoding, val)
        return (
            f"{p[0]:02x}:{p[1]:02x}:{p[2]:02x}:{p[5]:02x}:{p[6]:02x}:{p[7]:02x}/{p[8]}"
        )

    def i2repr(self, pkt, x):
        return self.i2h(pkt, x)



from .layers import PTPv2


def matching_sync(sync, fup):
    matching_portIds = sync.sourcePortIdentity == fup.sourcePortIdentity
    matching_seqIds = sync.sequenceId == fup.sequenceId
    return matching_portIds and matching_seqIds


def matching_pdelay(req, resp):
    matching_portIds = req.sourcePortIdentity == resp.requestingPortIdentity
    matching_seqIds = req.sequenceId == resp.sequenceId
    return matching_portIds and matching_seqIds


class MatchedList:
    def __init__(self, packets=[]):
        self._sync = []
        self._pdelay = []

        self._unmatched_sync = []
        self._unmatched_fup = []
        self._unmatched_pdreq = []
        self._unmatched_pdresp = []
        self._unmatched_pdresp_fup = []

        self.add(packets)

    def add(self, pkt):
        if type(pkt) is PTPv2:
            self._add_dispatch(pkt)
        elif iter(pkt):
            for p in pkt:
                self._add_dispatch(p)

    def _add_dispatch(self, p):
        if p.messageType == 0x0:
            self._add_sync(p)
        elif p.messageType == 0x2:
            self._add_pdelay_request(p)
        elif p.messageType == 0x3:
            self._add_pdelay_response(p)
        elif p.messageType == 0x8:
            self._add_followup(p)
        elif p.messageType == 0xA:
            self._add_pdelay_response_followup(p)

    def _add_sync(self, sync):
        for i, fup in enumerate(self._unmatched_fup):
            if matching_sync(sync, fup):
                self._sync.append((sync, self._unmatched_fup.pop(i)))
                return
        self._unmatched_sync.append(sync)

    def _add_followup(self, fup):
        for i, sync in enumerate(self._unmatched_sync):
            if matching_sync(sync, fup):
                self._sync.append((self._unmatched_sync.pop(i), fup))
                return
        self._unmatched_fup.append(fup)

    def _add_pdelay_request(self, req):
        for i, resp in enumerate(self._unmatched_pdresp):
            if matching_pdelay(req, resp):
                for j, respfup in enumerate(self._unmatched_pdresp_fup):
                    if matching_pdelay(req, respfup):
                        self._pdelay.append(
                            (
                                req,
                                self._unmatched_pdresp.pop(i),
                                self._unmatched_pdresp_fup.pop(j),
                            )
                        )
                        return
        self._unmatched_pdreq.append(req)

    def _add_pdelay_response(self, resp):
        for i, req in enumerate(self._unmatched_pdreq):
            if matching_pdelay(req, resp):
                for j, respfup in enumerate(self._unmatched_pdresp_fup):
                    if matching_pdelay(req, respfup):
                        self._pdelay.append(
                            (
                                self._unmatched_pdreq.pop(i),
                                resp,
                                self._unmatched_pdresp_fup.pop(j),
                            )
                        )
                        return
        self._unmatched_pdresp.append(resp)

    def _add_pdelay_response_followup(self, resp_fup):
        for i, req in enumerate(self._unmatched_pdreq):
            if matching_pdelay(req, resp_fup):
                for j, resp in enumerate(self._unmatched_pdresp):
                    if matching_pdelay(req, resp):
                        self._pdelay.append(
                            (
                                self._unmatched_pdreq.pop(i),
                                self._unmatched_pdresp.pop(j),
                                resp_fup,
                            )
                        )
                        return
        self._unmatched_pdresp_fup.append(resp_fup)

    @property
    def sync(self):
        return self._sync

    @property
    def pdelay(self):
        return self._pdelay

    @property
    def unmatched(self):
        return {
            "sync": self._unmatched_sync,
            "followup": self._unmatched_fup,
            "pdelay_req": self._unmatched_pdreq,
            "pdelay_resp": self._unmatched_pdresp,
            "pdelay_resp_followup": self._unmatched_pdresp_fup,
        }

    def __repr__(self):
        return f"<MatchedList sync:{len(self.sync)} pdelay:{len(self.pdelay)}>"
